from resolwe.process import *


class PythonProcessDataIdBySlug(Process):
    """The process is used for testing get_data_id_by_slug."""
    slug = 'test-python-process-data-id-by-slug'
    name = "Test Python Process Data ID by Slug"
    version = '1.0.0'
    process_type = 'data:python:dataidbyslug'
    requirements = {
        'executor': {
            'docker': {
                'image': 'resolwe/base:ubuntu-18.04',
            }
        },
        'resources': {
            'network': True,
        }
    }

    class Input:
        """Input fields."""
        slug = StringField(label="Slug")

    class Output:
        data_id = IntegerField(label="Data ID")

    def run(self, inputs, outputs):
        data_id = self.get_data_id_by_slug(inputs.slug)
        outputs.data_id = data_id



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-23 09:08
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0035_process_entity_new'),
    ]

    operations = [
        migrations.AddField(
            model_name='data',
            name='collection',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='data2', to='flow.Collection'),
        ),
        migrations.AddField(
            model_name='data',
            name='entity2',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='data2', to='flow.Entity'),
        ),
        migrations.AddField(
            model_name='entity',
            name='collection',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='entities2', to='flow.Collection'),
        ),
    ]



# Generated by Django 2.2 on 2019-04-24 07:21

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0033_move_purged'),
    ]

    operations = [
        migrations.AlterField(
            model_name='data',
            name='process_error',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=list, size=None),
        ),
        migrations.AlterField(
            model_name='data',
            name='process_info',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=list, size=None),
        ),
        migrations.AlterField(
            model_name='data',
            name='process_warning',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=list, size=None),
        ),
    ]



# pylint: disable=missing-docstring
from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS, connections
from django.test.utils import CaptureQueriesContext
from django.urls import reverse

from guardian.conf.settings import ANONYMOUS_USER_NAME
from guardian.models import UserObjectPermission
from guardian.shortcuts import assign_perm
from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.models import Collection, Data, DataDependency, DataLocation, DescriptorSchema, Entity, Process
from resolwe.flow.views import CollectionViewSet, DataViewSet, EntityViewSet, ProcessViewSet
from resolwe.test import ResolweAPITestCase, TestCase

factory = APIRequestFactory()  # pylint: disable=invalid-name


MESSAGES = {
    'NOT_FOUND': 'Not found.',
    'NO_PERMS': 'You do not have permission to perform this action.',
}


class TestDataViewSetCase(TestCase):
    def setUp(self):
        super().setUp()

        self.data_viewset = DataViewSet.as_view(actions={
            'get': 'list',
            'post': 'create',
        })
        self.duplicate_viewset = DataViewSet.as_view(actions={
            'post': 'duplicate',
        })
        self.data_detail_viewset = DataViewSet.as_view(actions={
            'get': 'retrieve',
            'patch': 'partial_update',
        })
        self.parents_viewset = DataViewSet.as_view(actions={
            'get': 'parents',
        })
        self.children_viewset = DataViewSet.as_view(actions={
            'get': 'children',
        })

        self.collection = Collection.objects.create(contributor=self.contributor)

        self.proc = Process.objects.create(
            type='data:test:process',
            slug='test-process',
            version='1.0.0',
            contributor=self.contributor,
            entity_type='test-schema',
            entity_descriptor_schema='test-schema',
            input_schema=[{'name': 'input_data', 'type': 'data:test:', 'required': False}],
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            slug='test-schema',
            version='1.0.0',
            contributor=self.contributor,
        )

        assign_perm('view_collection', self.contributor, self.collection)
        assign_perm('add_collection', self.contributor, self.collection)
        assign_perm('view_process', self.contributor, self.proc)
        assign_perm('view_descriptorschema', self.contributor, self.descriptor_schema)

    def test_prefetch(self):
        request = factory.get('/', '', format='json')
        force_authenticate(request, self.contributor)

        for _ in range(10):
            Data.objects.create(contributor=self.contributor, process=self.proc)

        # Check prefetch. The number of queries without prefetch depends
        # on the number of Data objects. With prefetch 56 queries,
        # without prefetch 73 queries. Python 2 and 3 have slightly
        # different number of queries, so we set a loose constraint in test.
        conn = connections[DEFAULT_DB_ALIAS]
        with CaptureQueriesContext(conn) as captured_queries:
            response = self.data_viewset(request)
            self.assertEqual(len(response.data), 10)
            self.assertLess(len(captured_queries), 90)

    def test_descriptor_schema(self):
        # Descriptor schema can be assigned by slug.
        data = {
            'process': {'slug': 'test-process'},
            'descriptor_schema': {'slug': 'test-schema'},
        }
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        self.data_viewset(request)

        data = Data.objects.latest()
        self.assertEqual(data.descriptor_schema, self.descriptor_schema)

        # Descriptor schema can be assigned by id.
        data = {
            'process': {'slug': 'test-process'},
            'descriptor_schema': {'slug': self.descriptor_schema.pk},
        }
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        self.data_viewset(request)

        data = Data.objects.latest()
        self.assertEqual(data.descriptor_schema, self.descriptor_schema)

    def test_use_latest_with_perm(self):
        Process.objects.create(
            type='test:process',
            name='Test process',
            slug='test-process',
            version='2.0.0',
            contributor=self.contributor,
        )
        DescriptorSchema.objects.create(
            name='Test schema',
            slug='test-schema',
            version='2.0.0',
            contributor=self.contributor,
        )

        data = {
            'process': {'slug': 'test-process'},
            'descriptor_schema': {'slug': 'test-schema'},
        }
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        self.data_viewset(request)

        data = Data.objects.latest()
        # Check that older versions are user if user doesn't have permissions on the latest
        self.assertEqual(data.process, self.proc)
        self.assertEqual(data.descriptor_schema, self.descriptor_schema)

    def test_public_create(self):
        assign_perm('view_process', AnonymousUser(), self.proc)

        data = {'process': {'slug': 'test-process'}}
        request = factory.post('/', data, format='json')
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Data.objects.count(), 1)

        data = Data.objects.latest()
        self.assertEqual(data.contributor.username, ANONYMOUS_USER_NAME)
        self.assertEqual(data.process.slug, 'test-process')

    def test_inherit_permissions(self):
        data_ctype = ContentType.objects.get_for_model(Data)
        entity_ctype = ContentType.objects.get_for_model(Entity)

        assign_perm('view_collection', self.user, self.collection)
        assign_perm('add_collection', self.user, self.collection)

        post_data = {
            'process': {'slug': 'test-process'},
            'collection': {'id': self.collection.pk},
        }
        request = factory.post('/', post_data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        data = Data.objects.last()
        entity = Entity.objects.last()

        self.assertTrue(self.user.has_perm('view_data', data))
        self.assertTrue(self.user.has_perm('view_entity', entity))
        self.assertTrue(self.user.has_perm('add_entity', entity))
        self.assertEqual(UserObjectPermission.objects.filter(content_type=data_ctype, user=self.user).count(), 1)
        self.assertEqual(UserObjectPermission.objects.filter(content_type=entity_ctype, user=self.user).count(), 2)

        # Add some permissions and run another process in same entity.
        assign_perm('edit_collection', self.user, self.collection)
        assign_perm('share_entity', self.user, entity)

        post_data = {
            'process': {'slug': 'test-process'},
            'collection': {'id': self.collection.pk},
            'input': {'input_data': data.pk},
        }
        request = factory.post('/', post_data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        data_2 = Data.objects.last()
        self.assertTrue(self.user.has_perm('view_data', data_2))
        self.assertTrue(self.user.has_perm('edit_data', data_2))
        self.assertTrue(self.user.has_perm('share_data', data_2))
        self.assertEqual(UserObjectPermission.objects.filter(content_type=data_ctype, user=self.user).count(), 4)
        self.assertEqual(UserObjectPermission.objects.filter(content_type=entity_ctype, user=self.user).count(), 3)

    def test_handle_entity(self):
        data = {'process': {'slug': 'test-process'}, 'collection': {'id': self.collection.pk}}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        # Test that one Entity was created and that it was added to the same collection as Data object.
        self.assertEqual(Entity.objects.count(), 1)
        self.assertEqual(Entity.objects.first().collection.pk, self.collection.pk)

    def test_collections_fields(self):
        # Create data object.
        data = {'process': {'slug': 'test-process'}, 'collection': {'id': self.collection.pk}}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_viewset(request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = Data.objects.last()
        entity = Entity.objects.last()

        # Ensure collection/entity are present
        request = factory.get('/', '', format='json')
        force_authenticate(request, self.contributor)
        response = self.data_viewset(request)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 1)
        self.assertIn('collection', response.data[0].keys())
        self.assertIn('entity', response.data[0].keys())

        # Check that query returns the correct collection ids.
        request = factory.get('/', '', format='json')
        force_authenticate(request, self.contributor)
        response = self.data_detail_viewset(request, pk=data.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['collection']['id'], self.collection.pk)
        self.assertEqual(response.data['entity']['id'], entity.pk)

    def test_collection_unassigned(self):
        # Data can be removed from collection through the api.

        # Create data object.
        data = {'process': {'slug': 'test-process'}, 'collection': {'id': self.collection.pk}}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_viewset(request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = Data.objects.last()
        entity = Entity.objects.last()

        self.assertEqual(data.collection.id, self.collection.id)
        self.assertEqual(data.entity.id, entity.id)
        self.assertEqual(entity.collection.id, self.collection.id)

        # Assign collection to None
        data.entity = None
        data.save()
        request = factory.patch('/', {'collection': {'id': None}}, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_detail_viewset(request, pk=data.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = Data.objects.last()
        self.assertEqual(data.collection, None)

    def test_change_collection(self):
        # Create data object. Note that an entity is created as well.
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )

        # Move data to some collection
        request = factory.patch('/', {'collection': {'id': self.collection.pk}}, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_detail_viewset(request, pk=data.pk)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.data['collection'][0],
            'If Data is in entity, you can only move it to another collection by moving entire entity.',
        )

        # But moving the data when it is not in entity is OK.
        data.entity = None
        data.save()
        request = factory.patch('/', {'collection': {'id': self.collection.pk}}, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_detail_viewset(request, pk=data.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data.refresh_from_db()
        self.assertEqual(data.collection, self.collection)

    def test_process_is_active(self):
        # Do not allow creating data of inactive processes
        Process.objects.filter(slug='test-process').update(is_active=False)
        data = {'process': {'slug': 'test-process'}}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_viewset(request)
        self.assertEqual(response.status_code, 400)

    def test_duplicate(self):
        data = Data.objects.create(contributor=self.contributor, process=self.proc)
        assign_perm("view_data", self.contributor, data)
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)
        data.status = Data.STATUS_DONE
        data.save()

        request = factory.post(reverse('resolwe-api:data-duplicate'), {'ids': [data.id]}, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)

        duplicate = Data.objects.get(id=response.data[0]['id'])
        self.assertTrue(duplicate.is_duplicate())

    def test_duplicate_not_auth(self):
        request = factory.post(reverse('resolwe-api:data-duplicate'), format='json')
        response = self.duplicate_viewset(request)

        self.assertEqual(response.data['detail'], MESSAGES['NOT_FOUND'])

    def test_duplicate_wrong_parameters(self):
        request = factory.post(reverse('resolwe-api:data-duplicate'), format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)
        self.assertEqual(response.data['detail'], "`ids` parameter is required")

        request = factory.post(reverse('resolwe-api:data-duplicate'), {'ids': 1}, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)
        self.assertEqual(response.data['detail'], "`ids` parameter not a list")

        request = factory.post(reverse('resolwe-api:data-duplicate'), {'ids': []}, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)
        self.assertEqual(response.data['detail'], "`ids` parameter is empty")

        request = factory.post(reverse('resolwe-api:data-duplicate'), {'ids': ['a']}, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)
        self.assertEqual(response.data['detail'], "`ids` parameter contains non-integers")

        request = factory.post(reverse('resolwe-api:data-duplicate'), {'ids': [0]}, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)
        self.assertEqual(response.data['detail'], "Data objects with the following ids not found: 0")

    def test_parents_children(self):
        parent = Data.objects.create(contributor=self.contributor, process=self.proc)
        child_1 = Data.objects.create(contributor=self.contributor, process=self.proc)
        child_2 = Data.objects.create(contributor=self.contributor, process=self.proc)

        DataDependency.objects.create(parent=parent, child=child_1, kind=DataDependency.KIND_IO)
        DataDependency.objects.create(parent=parent, child=child_2, kind=DataDependency.KIND_IO)

        assign_perm('view_data', self.user, parent)
        assign_perm('view_data', self.user, child_1)

        request = factory.get('/', format='json')
        force_authenticate(request, self.user)
        response = self.children_viewset(request, pk=parent.pk)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['id'], child_1.pk)

        request = factory.get('/', format='json')
        force_authenticate(request, self.user)
        response = self.parents_viewset(request, pk=child_1.pk)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['id'], parent.pk)


class TestCollectionViewSetCase(TestCase):
    def setUp(self):
        super().setUp()

        self.checkslug_viewset = CollectionViewSet.as_view(actions={
            'get': 'slug_exists',
        })
        self.add_data_viewset = CollectionViewSet.as_view(actions={
            'post': 'add_data',
        })
        self.remove_data_viewset = CollectionViewSet.as_view(actions={
            'post': 'remove_data',
        })
        self.duplicate_viewset = CollectionViewSet.as_view(actions={
            'post': 'duplicate',
        })
        self.collection_detail_viewset = CollectionViewSet.as_view(actions={
            'get': 'retrieve',
            'put': 'update',
            'patch': 'partial_update',
            'delete': 'destroy',
        })
        self.collection_list_viewset = CollectionViewSet.as_view(actions={
            'get': 'list',
            'post': 'create',
        })

        self.detail_url = lambda pk: reverse('resolwe-api:collection-detail', kwargs={'pk': pk})

    def _create_data(self):
        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        return Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=process,
        )

    def _create_entity(self):
        return Entity.objects.create(
            name='Test entity',
            contributor=self.contributor,
        )

    def test_set_descriptor_schema(self):
        d_schema = DescriptorSchema.objects.create(slug="new-schema", name="New Schema", contributor=self.contributor)

        data = {
            'name': 'Test collection',
            'descriptor_schema': {'slug': 'new-schema'},
        }

        request = factory.post('/', data=data, format='json')
        force_authenticate(request, self.admin)
        self.collection_list_viewset(request)

        self.assertEqual(Collection.objects.count(), 1)
        self.assertEqual(Collection.objects.first().descriptor_schema, d_schema)

    def test_change_descriptor_schema(self):
        collection = Collection.objects.create(slug="collection1", name="Collection 1", contributor=self.contributor)
        d_schema = DescriptorSchema.objects.create(slug="new-schema", name="New Schema", contributor=self.contributor)

        # For updates, id must be used.
        data = {'descriptor_schema': {'id': d_schema.pk}}
        request = factory.patch(self.detail_url(collection.pk), data=data, format='json')
        force_authenticate(request, self.admin)
        self.collection_detail_viewset(request, pk=collection.pk)

        collection.refresh_from_db()
        self.assertEqual(collection.descriptor_schema, d_schema)

    def test_change_slug(self):
        collection1 = Collection.objects.create(name="Collection", contributor=self.contributor)
        collection2 = Collection.objects.create(name="Collection", contributor=self.contributor)
        self.assertEqual(collection1.slug, 'collection')
        self.assertEqual(collection2.slug, 'collection-2')

        request = factory.patch(self.detail_url(collection1.pk), {'name': 'Collection', 'slug': None}, format='json')
        force_authenticate(request, self.admin)
        response = self.collection_detail_viewset(request, pk=collection1.pk)
        self.assertEqual(response.data['slug'], 'collection')

        request = factory.patch(self.detail_url(collection2.pk), {'slug': 'collection-3'}, format='json')
        force_authenticate(request, self.admin)
        response = self.collection_detail_viewset(request, pk=collection2.pk)
        self.assertEqual(response.data['slug'], 'collection-3')

        request = factory.patch(self.detail_url(collection2.pk), {'slug': 'collection'}, format='json')
        force_authenticate(request, self.admin)
        response = self.collection_detail_viewset(request, pk=collection2.pk)
        self.assertContains(response, 'already taken', status_code=400)

    def test_check_slug(self):
        Collection.objects.create(slug="collection1", name="Collection 1", contributor=self.admin)

        # unauthorized
        request = factory.get('/', {'name': 'collection1'}, format='json')
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.status_code, 401)
        self.assertEqual(resp.data, None)

        # existing slug
        request = factory.get('/', {'name': 'collection1'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, True)

        # existing slug - iexact
        request = factory.get('/', {'name': 'Collection1'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, True)

        # non-existing slug
        request = factory.get('/', {'name': 'new-collection'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, False)

        # bad query parameter
        request = factory.get('/', {'bad': 'parameter'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.status_code, 400)

    def test_delete(self):
        collection = Collection.objects.create(
            name="Test collection",
            contributor=self.contributor,
        )

        data_1, data_2 = self._create_data(), self._create_data()
        entity_1, entity_2 = self._create_entity(), self._create_entity()

        collection.data.add(data_1, data_2)
        collection.entity_set.add(entity_1, entity_2)

        assign_perm("view_collection", self.user, collection)
        assign_perm("edit_collection", self.user, collection)
        assign_perm("view_data", self.user, data_1)
        assign_perm("view_data", self.user, data_2)
        assign_perm("edit_data", self.user, data_1)
        assign_perm("view_entity", self.user, entity_1)
        assign_perm("view_entity", self.user, entity_2)
        assign_perm("edit_entity", self.user, entity_1)

        request = factory.delete(self.detail_url(collection.pk))
        force_authenticate(request, self.user)
        self.collection_detail_viewset(request, pk=collection.pk)

        # All containing objects are deleted, regardless of their permission.
        self.assertFalse(Data.objects.filter(pk=data_1.pk).exists())
        self.assertFalse(Data.objects.filter(pk=data_2.pk).exists())
        self.assertFalse(Entity.objects.filter(pk=entity_1.pk).exists())
        self.assertFalse(Entity.objects.filter(pk=entity_2.pk).exists())

    def test_duplicate(self):
        request = factory.post('/', {}, format='json')
        force_authenticate(request, self.contributor)
        self.collection_list_viewset(request)

        collection = Collection.objects.first()

        request = factory.post(reverse('resolwe-api:collection-duplicate'), {'ids': [collection.id]}, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)

        duplicate = Collection.objects.get(id=response.data[0]['id'])
        self.assertTrue(duplicate.is_duplicate())

    def test_duplicate_not_auth(self):
        request = factory.post(reverse('resolwe-api:collection-duplicate'), format='json')
        response = self.duplicate_viewset(request)

        self.assertEqual(response.data['detail'], MESSAGES['NOT_FOUND'])


class ProcessTestCase(ResolweAPITestCase):
    def setUp(self):
        self.resource_name = 'process'
        self.viewset = ProcessViewSet

        super().setUp()

    def test_create_new(self):
        post_data = {
            'name': 'Test process',
            'slug': 'test-process',
            'type': 'data:test:',
        }

        # Normal user is not allowed to create new processes.
        resp = self._post(post_data, self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

        # Superuser can create process.
        resp = self._post(post_data, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

    def test_is_active(self):
        post_data = {
            'name': 'Test process',
            'slug': 'test-process',
            'type': 'data:test:',
            'is_active': False,
        }

        # is_active can not be set through API and is True by default
        response = self._post(post_data, self.admin)
        self.assertTrue(response.data['is_active'])

        # is_active should not be changed through API
        process_id = response.data['id']
        response = self._patch(process_id, {'is_active': False}, self.admin)
        self.assertEqual(response.status_code, 405)  # PATCH not allowed on process


class EntityViewSetTest(TestCase):
    def setUp(self):
        super().setUp()

        self.collection = Collection.objects.create(name="Test Collection", contributor=self.contributor)
        self.collection2 = Collection.objects.create(name="Test Collection 2", contributor=self.contributor)
        self.entity = Entity.objects.create(name="Test entity", contributor=self.contributor)
        process = Process.objects.create(name="Test process", contributor=self.contributor)
        self.data = Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=process,
            status=Data.STATUS_DONE
        )
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(self.data)
        self.data_2 = Data.objects.create(
            name="Test data 2",
            contributor=self.contributor,
            process=process,
            status=Data.STATUS_DONE
        )
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(self.data_2)

        # another Data object to make sure that other objects are not processed
        data = Data.objects.create(name="Dummy data", contributor=self.contributor, process=process)
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        self.entity.data.add(self.data)
        self.entity.collection = self.collection2
        self.entity.save()

        assign_perm('add_collection', self.contributor, self.collection)
        assign_perm('add_entity', self.contributor, self.entity)
        assign_perm('view_collection', self.contributor, self.collection)
        assign_perm('view_collection', self.contributor, self.collection2)
        assign_perm('view_entity', self.contributor, self.entity)

        self.entityviewset = EntityViewSet()

        self.duplicate_viewset = EntityViewSet.as_view(actions={
            'post': 'duplicate',
        })
        self.move_to_collection_viewset = EntityViewSet.as_view(actions={
            'post': 'move_to_collection',
        })
        self.entity_detail_viewset = EntityViewSet.as_view(actions={
            'get': 'retrieve',
            'put': 'update',
            'patch': 'partial_update',
            'delete': 'destroy',
        })
        self.entity_list_viewset = EntityViewSet.as_view(actions={
            'get': 'list',
            'post': 'create',
        })

        self.detail_url = lambda pk: reverse('resolwe-api:entity-detail', kwargs={'pk': pk})

    def _create_data(self):
        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        return Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=process,
        )

    def test_list_filter_collection(self):
        request = factory.get('/', {}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 1)

        request = factory.get('/', {'collection': 999999}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 0)

        request = factory.get('/', {'collection': self.collection.pk}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 0)

        request = factory.get('/', {'collection': self.collection2.pk}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 1)

    def test_change_collection(self):
        self.collection.tags = ['test:tag']
        self.collection.save()
        self.data.collection = self.collection2
        self.data.save()
        assign_perm('edit_entity', self.contributor, self.entity)

        request_data = {'collection': {'id': self.collection.pk}}
        request = factory.patch('/', request_data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_detail_viewset(request, pk=self.entity.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.entity.refresh_from_db()
        self.data.refresh_from_db()
        self.assertEqual(self.data.entity.id, self.entity.id)
        self.assertEqual(self.data.collection.id, self.collection.id)
        self.assertEqual(self.entity.collection.id, self.collection.id)
        self.assertEqual(self.entity.tags, self.collection.tags)
        self.assertEqual(self.data.tags, self.collection.tags)

    def test_change_collection_to_none(self):
        assign_perm('edit_entity', self.contributor, self.entity)
        self.data.collection = self.collection2
        self.data.save()

        request_data = {'collection': {'id': None}}
        request = factory.patch('/', request_data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_detail_viewset(request, pk=self.entity.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.entity.refresh_from_db()
        self.data.refresh_from_db()
        self.assertEqual(self.data.entity.id, self.entity.id)
        self.assertEqual(self.data.collection, None)
        self.assertEqual(self.entity.collection, None)

    def test_move_to_collection(self):
        entity = Entity.objects.create(contributor=self.contributor)
        assign_perm('view_entity', self.contributor, entity)
        data = self._create_data()
        assign_perm('view_data', self.contributor, data)
        entity.data.add(data)

        source_collection = Collection.objects.create(contributor=self.contributor)
        assign_perm('view_collection', self.contributor, source_collection)
        assign_perm('add_collection', self.contributor, source_collection)
        entity.collection = source_collection
        entity.save()
        data.collection = source_collection
        data.save()

        destination_collection = Collection.objects.create(contributor=self.contributor)
        assign_perm('view_collection', self.contributor, destination_collection)
        assign_perm('add_collection', self.contributor, destination_collection)

        request = factory.post(reverse('resolwe-api:entity-move-to-collection'), {
            'ids': [entity.id],
            'source_collection': source_collection.id,
            'destination_collection': destination_collection.id,
        }, format='json')
        force_authenticate(request, self.contributor)

        self.assertEqual(source_collection.entity_set.count(), 1)
        self.assertEqual(source_collection.data.count(), 1)
        self.assertEqual(destination_collection.entity_set.count(), 0)
        self.assertEqual(destination_collection.data.count(), 0)

        self.move_to_collection_viewset(request)

        self.assertEqual(source_collection.entity_set.count(), 0)
        self.assertEqual(source_collection.data.count(), 0)
        self.assertEqual(destination_collection.entity_set.count(), 1)
        self.assertEqual(destination_collection.entity_set.first().id, entity.id)
        self.assertEqual(destination_collection.data.first().id, data.id)

    def test_delete(self):
        entity = Entity.objects.create(
            name="Test entity",
            contributor=self.contributor,
        )

        data_1, data_2 = self._create_data(), self._create_data()

        entity.data.add(data_1, data_2)

        assign_perm('view_entity', self.user, entity)
        assign_perm('edit_entity', self.user, entity)
        assign_perm('view_data', self.user, data_1)
        assign_perm('view_data', self.user, data_2)
        assign_perm('edit_data', self.user, data_1)

        request = factory.delete(self.detail_url(entity.pk))
        force_authenticate(request, self.user)
        self.entity_detail_viewset(request, pk=entity.pk)

        # If user has edit permison on entity, all containing objects
        # are deleted, regardless of their permission.
        self.assertFalse(Data.objects.filter(pk=data_1.pk).exists())
        self.assertFalse(Data.objects.filter(pk=data_2.pk).exists())

    def test_duplicate(self):
        entity = Entity.objects.first()
        collection = Collection.objects.create(contributor=self.contributor)
        assign_perm('add_collection', self.contributor, collection)
        collection.entity_set.add(entity)
        data = entity.data.all()
        for datum in data:
            assign_perm('view_data', self.contributor, datum)
        collection.data.add(*data)

        request = factory.post(reverse('resolwe-api:entity-duplicate'), {'ids': [entity.id]}, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)

        duplicate = Entity.objects.get(id=response.data[0]['id'])
        self.assertTrue(duplicate.is_duplicate())
        self.assertEqual(collection.entity_set.count(), 1)
        self.assertEqual(collection.data.count(), 1)

        request = factory.post(reverse('resolwe-api:entity-duplicate'), {
            'ids': [entity.id],
            'inherit_collection': True
        }, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)

        self.assertEqual(collection.entity_set.count(), 2)
        self.assertEqual(collection.data.count(), 2)

        # Assert collection membership.
        collection_without_perm = Collection.objects.create(contributor=self.contributor)
        collection_without_perm.entity_set.add(entity)
        collection_without_perm.data.add(*entity.data.all())

        request = factory.post(reverse('resolwe-api:entity-duplicate'), {
            'ids': [entity.id],
            'inherit_collection': True
        }, format='json')
        force_authenticate(request, self.contributor)
        response = self.duplicate_viewset(request)

        self.assertEqual(collection_without_perm.entity_set.count(), 1)
        self.assertEqual(collection_without_perm.data.count(), 1)

    def test_duplicate_not_auth(self):
        request = factory.post(reverse('resolwe-api:entity-duplicate'), format='json')
        response = self.duplicate_viewset(request)

        self.assertEqual(response.data['detail'], MESSAGES['NOT_FOUND'])



# pylint: disable=missing-docstring
import os
import sys
import unittest

from django.contrib.auth.models import AnonymousUser
from django.test import LiveServerTestCase, override_settings

from guardian.shortcuts import assign_perm

from resolwe.flow.models import Data, Entity, Process
from resolwe.test import ProcessTestCase, tag_process, with_docker_executor, with_resolwe_host

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), 'processes')
WORKFLOWS_DIR = os.path.join(os.path.dirname(__file__), 'workflows')
DESCRIPTORS_DIR = os.path.join(os.path.dirname(__file__), 'descriptors')
FILES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')


class PythonProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(processes_paths=[PROCESSES_DIR, WORKFLOWS_DIR], descriptors_paths=[DESCRIPTORS_DIR])
        self.files_path = FILES_PATH

    @with_docker_executor
    def test_registration(self):
        process = Process.objects.get(slug='test-python-process')

        self.assertEqual(process.slug, 'test-python-process')
        self.assertEqual(process.name, 'Test Python Process')
        self.assertEqual(process.version, '0.1.2')
        self.assertEqual(process.type, 'data:python:')
        self.assertEqual(process.category, 'analyses:')
        self.assertEqual(process.scheduling_class, Process.SCHEDULING_CLASS_BATCH)
        self.assertEqual(process.persistence, Process.PERSISTENCE_CACHED)
        self.assertEqual(process.description, 'This is a process description.')
        self.assertEqual(process.data_name, 'Foo: {{input_data | name}}')
        self.assertEqual(process.entity_type, 'sample')
        self.assertEqual(process.entity_descriptor_schema, 'sample')
        self.assertEqual(process.entity_input, 'input_data')
        self.assertEqual(process.requirements, {
            'expression-engine': 'jinja',
            'executor': {
                'docker': {
                    'image': 'resolwe/base:ubuntu-18.04',
                }
            }
        })

        for field in process.input_schema:
            if field['name'] == 'my_group':
                self.assertEqual(field['group'][0]['name'], 'foo')
                self.assertEqual(field['group'][1]['name'], 'bar')
                break
        else:
            self.fail("Field my_group not found in test-python-process")

        # Make sure that process with inheritance from `module.Class` is also registered.
        process = Process.objects.get(slug='test-python-process-2')

    @with_docker_executor
    @tag_process('test-python-process', 'test-save-file', 'entity-process')
    def test_python_process(self):
        with self.preparation_stage():
            input_data = self.run_process('test-save-file', {'input_file': 'testfile.txt'})
            input_data.name = "bar"
            input_data.save()

            input_entity = self.run_process('entity-process')

            storage = input_data.storages.create(
                name="storage",
                contributor=self.user,
                json={'value': 42}
            )

        data = self.run_process('test-python-process', {
            'my_field': "bar",
            'my_list': ["one", "two", "three"],
            'bar': input_data.pk,
            'url': {'url': "https://www.genialis.com"},
            'input_data': input_data.pk,
            'input_entity_data': input_entity.pk,
            'integer': 42,
            'my_float': 0.42,
            'my_json': storage.pk,
            'my_group': {
                'bar': 'my string',
                'foo': 21,
            }
        })

        self.assertFields(data, 'string_output', 'OK')
        self.assertFields(data, 'list_string_output', ['foo', 'bar'])
        self.assertFile(data, 'file_output', 'testfile.txt')
        self.assertFiles(data, 'list_file_output', ['testfile.txt', 'testfile2.txt'])

        # Non-deterministic output.
        del data.output['dir_output']['size']
        del data.output['dir_output']['total_size']

        self.assertEqual(data.output['dir_output'], {'dir': 'test/'})
        self.assertEqual(data.output['input_entity_name'], 'Data with entity')
        self.assertEqual(data.output['docker_image'], 'resolwe/base:ubuntu-18.04')
        self.assertEqual(data.name, 'Foo: bar')

        entity = Entity.objects.get(data=data)
        self.assertEqual(entity.data.first(), data)

    @with_docker_executor
    @tag_process('test-python-process-json')
    def test_python_process_json(self):
        """Test that data object with json output can be given as input."""
        with self.preparation_stage():
            input_data = self.run_process('test-output-json')

        self.run_process('test-python-process-json', {
            'data': input_data.pk,
        })

    @with_docker_executor
    @tag_process('test-non-required-data-inputs')
    def test_non_required_data_input(self):
        """Test workflow with non-required data inputs"""
        with self.preparation_stage():
            input_data = self.run_process('test-output-json')

        self.run_process('test-non-required-data-inputs', {
            'data': input_data.pk,
        })

        data = Data.objects.get(process__slug='test-python-process-json')
        self.assertEqual(data.status, 'OK')

    @with_docker_executor
    @tag_process('test-python-process-error')
    def test_error(self):
        """Test process that raises exception"""
        data = self.run_process('test-python-process-error', assert_status=Data.STATUS_ERROR)
        self.assertEqual(data.process_error[0], 'Value error in ErrorProcess')

    @with_docker_executor
    @tag_process('test-python-process-file')
    def test_import_file(self):
        """Test import file"""
        inputs = {'src': 'testfile.txt'}
        data = self.run_process('test-python-process-file', inputs)
        self.assertEqual(data.output['dst']['file'], 'testfile.txt')
        self.assertEqual(data.output['dst']['size'], 15)


class PythonProcessRequirementsTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(processes_paths=[PROCESSES_DIR, WORKFLOWS_DIR], descriptors_paths=[DESCRIPTORS_DIR])
        self.files_path = FILES_PATH

    @with_docker_executor
    @tag_process('test-python-process-requirements')
    def test_defaults(self):
        data = self.run_process('test-python-process-requirements')
        self.assertEqual(data.output['cores'], 2)
        self.assertEqual(data.output['memory'], 4096)

    @with_docker_executor
    @override_settings(FLOW_PROCESS_MAX_CORES=1)
    @tag_process('test-python-process-requirements')
    def test_max_cores(self):
        data = self.run_process('test-python-process-requirements')
        self.assertEqual(data.output['cores'], 1)
        self.assertEqual(data.output['memory'], 4096)

    @with_docker_executor
    @override_settings(FLOW_PROCESS_RESOURCE_OVERRIDES={'memory': {'test-python-process-requirements': 2048}})
    @tag_process('test-python-process-requirements')
    def test_resource_override(self):
        data = self.run_process('test-python-process-requirements')
        self.assertEqual(data.output['cores'], 2)
        self.assertEqual(data.output['memory'], 2048)


class PythonProcessDataBySlugTest(ProcessTestCase, LiveServerTestCase):

    def setUp(self):
        super().setUp()
        self._register_schemas(processes_paths=[PROCESSES_DIR], descriptors_paths=[DESCRIPTORS_DIR])
        self.files_path = FILES_PATH

    @unittest.skipUnless(
        sys.platform.startswith('linux'),
        "Accessing live Resolwe host from a Docker container on non-Linux systems is not possible yet."
    )
    @with_resolwe_host
    @with_docker_executor
    @tag_process('test-python-process-data-id-by-slug', 'test-python-process-2')
    def test_process_data_by_slug(self):
        """Test that data object with json output can be given as input."""
        with self.preparation_stage():
            input_data = self.run_process('test-python-process-2')

        input_data = Data.objects.get(id=input_data.id)
        assign_perm("view_data", AnonymousUser(), input_data)

        data = self.run_process('test-python-process-data-id-by-slug', {
            'slug': input_data.slug,
        })

        self.assertEqual(data.output['data_id'], input_data.pk)



"""Resolwe collection serializer."""
from resolwe.flow.models import Collection, DescriptorSchema
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer
from .descriptor import DescriptorSchemaSerializer
from .fields import DictRelatedField


class CollectionSerializer(ResolweBaseSerializer):
    """Serializer for Collection objects."""

    settings = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)
    descriptor_schema = DictRelatedField(
        queryset=DescriptorSchema.objects.all(),
        serializer=DescriptorSchemaSerializer,
        allow_null=True,
        required=False,
    )

    class Meta:
        """CollectionSerializer Meta options."""

        model = Collection
        read_only_fields = (
            'created',
            'descriptor_dirty',
            'duplicated',
            'id',
            'modified',
        )
        update_protected_fields = (
            'contributor',
        )
        fields = read_only_fields + update_protected_fields + (
            'description',
            'descriptor',
            'descriptor_schema',
            'name',
            'settings',
            'slug',
            'tags',
        )



from resolwe import process
from resolwe.process import *


class EntityProcess(Process):
    slug = 'entity-process'
    name = "Entity process"
    data_name = "Data with entity"
    version = '1.0.0'
    process_type = 'data:entity'
    entity = {
        'type': 'sample',
    }

    class Output:
        list_string = ListField(StringField(), label="My list")
        optional = StringField("Optional output", required=False)

    def run(self, inputs, outputs):
        outputs.list_string = ["foo", "bar"]


class PythonProcess(Process):
    """This is a process description."""
    slug = 'test-python-process'
    name = "Test Python Process"
    version = '0.1.2'
    process_type = 'data:python'
    category = 'analyses'
    scheduling_class = SchedulingClass.BATCH
    persistence = Persistence.CACHED
    data_name = "Foo: {{input_data | name}}"
    entity = {
        'type': 'sample',
        'descriptor_schema': 'sample',
        'input': 'input_data',
    }
    requirements = {
        'expression-engine': 'jinja',
        'executor': {
            'docker': {
                'image': 'resolwe/base:ubuntu-18.04',
            }
        }
    }

    class Input:
        """Input fields."""
        my_field = StringField(label="My field")
        my_list = ListField(StringField(), label="My list")
        input_data = DataField('test:save', label="My input data")
        input_entity_data = DataField('entity', label="My entity data")
        bar = DataField(data_type='test:save', label="My bar")
        url = UrlField(UrlField.DOWNLOAD, label="My URL")
        integer = IntegerField(label="My integer")
        my_float = FloatField(label="My float")
        my_json = JsonField(label="Blah blah")
        my_optional = StringField(label="Optional", required=False, default='default value')
        my_optional_no_default = StringField(label="Optional no default", required=False)

        class MyGroup:
            foo = IntegerField(label="Foo")
            bar = StringField(label="Bar")
            group_optional_no_default = StringField(label="Group optional no default", required=False)

        my_group = GroupField(MyGroup, label="My group")

    class Output:
        string_output = StringField(label="My string output")
        list_string_output = ListField(StringField(), label="My list string output")
        file_output = FileField(label="My output")
        list_file_output = ListField(FileField(), label="My list output")
        dir_output = DirField(label="My output")
        input_entity_name = StringField(label="Input entity name")
        docker_image = StringField(label="Docker image")

    def run(self, inputs, outputs):
        print('All inputs are:', inputs)
        print('Input data:', inputs.input_data)
        print('Input data ID:', inputs.input_data.id)
        print('Input data file output:', inputs.input_data.saved_file.path)
        print('Input data type:', inputs.input_data.type)
        print('Input data descriptor:', inputs.input_data.descriptor)
        print('Group bar:', inputs.my_group.bar)
        print('Group foo:', inputs.my_group.foo)
        print('Entity name of the input:', inputs.input_entity_data.entity_name)
        print('Docker image:', self.requirements.executor.docker.image)

        if inputs.my_optional:
            print('My optional:', inputs.my_optional)

        if inputs.my_optional_no_default:
            raise AttributeError('inputs.my_optional_no_default should not exist.')

        if inputs.my_group.group_optional_no_default:
            raise AttributeError('inputs.my_group.group_optional_no_default should not exist.')

        if inputs.input_entity_data.optional:
            raise AttributeError('inputs.list_string_output.optional should not exist.')

        try:
            inputs.invalid_input
        except AttributeError as err:
            if 'Inputs have no field invalid_input' in str(err):
                pass

        try:
            inputs.input_entity_data.invalid_field
        except AttributeError as err:
            if 'DataField has no member invalid_field' in str(err):
                pass

        bar = Cmd['ls']['-l', '-a', '/'] | Cmd['grep']['python']
        print('hello world:\n', bar())

        cmd = Cmd['mkdir']['test']()
        cmd = (Cmd['echo']['"Some content"'] > 'test/testfile.txt')()
        cmd = (Cmd['echo']['"Some more content"'] > 'testfile2.txt')()

        outputs.file_output = 'test/testfile.txt'
        outputs.list_file_output = ['test/testfile.txt', 'testfile2.txt']
        outputs.dir_output = 'test/'
        outputs.input_entity_name = inputs.input_entity_data.entity_name
        outputs.docker_image = self.requirements.executor.docker.image
        outputs.string_output = 'OK'
        outputs.list_string_output = ['foo', 'bar']


class PythonProcess2(process.Process):
    """Inherit from 'module.Class'."""
    slug = 'test-python-process-2'
    name = "Test Python Process 2"
    version = '0.0.1'
    process_type = 'data:python'

    def run(self, inputs, outputs):
        pass


class ErrorProcess(Process):
    slug = 'test-python-process-error'
    name = "Test Python Process Error"
    version = '0.0.1'
    process_type = 'data:python:error'

    def run(self, inputs, outputs):
        raise ValueError('Value error in ErrorProcess')


class FileProcess(Process):
    slug = 'test-python-process-file'
    name = "Test Python Process File"
    version = '0.0.1'
    process_type = 'data:python:file'

    class Input:
        """Input fields."""
        src = FileField(label="Input file")

    class Output:
        """Input fields."""
        dst = FileField(label="Output file")

    def run(self, inputs, outputs):
        file_name = inputs.src.import_file()
        outputs.dst = file_name


class RequirementsProcess(Process):
    slug = 'test-python-process-requirements'
    name = "Test Python Process Requirements"
    version = '0.0.1'
    process_type = 'data:python:requirements'
    requirements = {
        'resources': {
            'cores': 2,
            'memory': 4096,
        },
    }

    class Output:
        """Input fields."""
        cores = IntegerField(label="Cores")
        memory = IntegerField(label="Memory")

    def run(self, inputs, outputs):
        outputs.cores = self.requirements['resources']['cores']
        outputs.memory = self.requirements['resources']['memory']

        print('Cores:', outputs.cores)
        print('Memory:', outputs.memory)



"""Elastic Search indexes for Data model."""
import elasticsearch_dsl as dsl

from resolwe.elastic.fields import Name, ProcessType
from resolwe.elastic.indices import BaseIndex

from ..models import Data
from .base import BaseDocument, BaseIndexMixin


class DataDocument(BaseDocument):
    """Document for data search."""

    started = dsl.Date()
    finished = dsl.Date()
    status = dsl.Keyword()
    process = dsl.Integer()
    process_type = ProcessType()
    # Keep backward compatibility.
    type = ProcessType()
    process_name = Name()
    tags = dsl.Keyword(multi=True)

    collection = dsl.Integer()
    entity = dsl.Integer()

    class Index:
        """Meta class for data search document."""

        name = 'data'


class DataIndex(BaseIndexMixin, BaseIndex):
    """Index for data objects used in ``DataDocument``."""

    queryset = Data.objects.all().prefetch_related(
        'process',
        'contributor'
    )
    object_type = Data
    document_class = DataDocument

    mapping = {
        'process': 'process.id',
        'process_name': 'process.name',
        'process_type': 'process.type',
        'type': 'process.type',
        'entity': 'entity_id',
        'collection': 'collection_id',
    }



# Generated by Django 2.2.1 on 2019-06-18 14:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0034_callable_defaults'),
    ]

    operations = [
        migrations.AddField(
            model_name='process',
            name='entity_always_create',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.2 on 2019-07-24 07:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0038_remove_m2m_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='data',
            name='collection',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='data', to='flow.Collection'),
        ),
        migrations.AlterField(
            model_name='data',
            name='entity',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='data', to='flow.Entity'),
        ),
        migrations.AlterField(
            model_name='entity',
            name='collection',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='flow.Collection'),
        ),
    ]



"""Resolwe Python process support."""
from .runtime import *

# XXX: ImportedFormat is available from resolwe-runtime-utils v2.0.0. Docker
# images with older version of resolwe-runtime-utils should skip the import.
try:
    from resolwe_runtime_utils import ImportedFormat
except ImportError:
    pass



"""Entity viewset."""
from rest_framework import exceptions
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.flow.models import Collection, Entity
from resolwe.flow.serializers import EntitySerializer
from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.permissions.utils import remove_permission, update_permission

from ..elastic_indexes import EntityDocument
from .collection import CollectionViewSet


class EntityViewSet(CollectionViewSet):
    """API view for entities."""

    serializer_class = EntitySerializer
    document_class = EntityDocument
    queryset = Entity.objects.prefetch_related('descriptor_schema', 'contributor')
    filtering_fields = CollectionViewSet.filtering_fields + ('descriptor_completed', 'collection', 'type')

    def _get_collection_for_user(self, collection_id, user):
        """Check that collection exists and user has `add` permission."""
        collection_query = Collection.objects.filter(pk=collection_id)
        if not collection_query.exists():
            raise exceptions.ValidationError('Collection id does not exist')

        collection = collection_query.first()
        if not user.has_perm('add_collection', obj=collection):
            if user.is_authenticated:
                raise exceptions.PermissionDenied()
            else:
                raise exceptions.NotFound()

        return collection

    def _get_entities(self, user, ids):
        """Return entities queryset based on provided entity ids."""
        queryset = get_objects_for_user(user, 'view_entity', Entity.objects.filter(id__in=ids))
        actual_ids = queryset.values_list('id', flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}" .format(', '.join(map(str, missing_ids)))
            )

        return queryset

    def set_content_permissions(self, user, obj, payload):
        """Apply permissions to data objects in ``Entity``."""
        # Data doesn't have "ADD" permission, so it has to be removed
        payload = remove_permission(payload, 'add')

        for data in obj.data.all():
            if user.has_perm('share_data', data):
                update_permission(data, payload)

    @action(detail=False, methods=['post'])
    def move_to_collection(self, request, *args, **kwargs):
        """Move samples from source to destination collection."""
        ids = self.get_ids(request.data)
        src_collection_id = self.get_id(request.data, 'source_collection')
        dst_collection_id = self.get_id(request.data, 'destination_collection')

        src_collection = self._get_collection_for_user(src_collection_id, request.user)
        dst_collection = self._get_collection_for_user(dst_collection_id, request.user)

        entity_qs = self._get_entities(request.user, ids)
        entity_qs.move_to_collection(src_collection, dst_collection)

        return Response()

    # NOTE: This can be deleted when DRF will support select_for_update
    #       on updates and ResolweUpdateModelMixin will use it.
    #       https://github.com/encode/django-rest-framework/issues/4675
    def update(self, request, *args, **kwargs):
        """Update an entity.

        Original queryset produces a temporary database table whose rows
        cannot be selected for an update. As a workaround, we patch
        get_queryset function to return only Entity objects without
        additional data that is not needed for the update.
        """
        orig_get_queryset = self.get_queryset

        def patched_get_queryset():
            """Patched get_queryset method."""
            entity_ids = orig_get_queryset().values_list('id', flat=True)
            return Entity.objects.filter(id__in=entity_ids)

        self.get_queryset = patched_get_queryset
        resp = super().update(request, *args, **kwargs)
        self.get_queryset = orig_get_queryset
        return resp

    @action(detail=False, methods=['post'])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Entity`` models."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        inherit_collection = request.data.get('inherit_collection', False)
        ids = self.get_ids(request.data)
        queryset = get_objects_for_user(request.user, 'view_entity', Entity.objects.filter(id__in=ids))
        actual_ids = queryset.values_list('id', flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(', '.join(map(str, missing_ids)))
            )

        duplicated = queryset.duplicate(contributor=request.user, inherit_collection=inherit_collection)

        serializer = self.get_serializer(duplicated, many=True)
        return Response(serializer.data)



""".. Ignore pydocstyle D400.

============
Flow Filters
============

"""
import django_filters as filters

from rest_framework.exceptions import ParseError

from .models import Collection, Data, DescriptorSchema, Entity, Process, Relation

NUMBER_LOOKUPS = [
    'exact',
    'in',
    'gt', 'gte', 'lt', 'lte',
    'isnull',
]
TEXT_LOOKUPS = [
    'exact', 'iexact',
    'contains', 'icontains',
    'in',
    'startswith', 'istartswith',
    'endswith', 'iendswith',
    'isnull',
]
DATE_LOOKUPS = [
    'exact',
    'gt', 'gte', 'lt', 'lte',
    'year', 'year__gt', 'year__gte', 'year__lt', 'year__lte',
    'month', 'month__gt', 'month__gte', 'month__lt', 'month__lte',
    'day', 'day__gt', 'day__gte', 'day__lt', 'day__lte',
    'isnull',

]
DATETIME_LOOKUPS = DATE_LOOKUPS + [
    'date',
    'time',
    'hour', 'hour__gt', 'hour__gte', 'hour__lt', 'hour__lte',
    'minute', 'minute__gt', 'minute__gte', 'minute__lt', 'minute__lte',
    'second', 'second__gt', 'second__gte', 'second__lt', 'second__lte',
]


class CheckQueryParamsMixin:
    """Custom query params validation."""

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return (
            'fields',
            'format',
            'limit',
            'offset',
            'ordering',
        )

    def validate_query_params(self):
        """Ensure no unsupported query params were used."""
        allowed_params = set(self.get_filters().keys())
        allowed_params.update(self.get_always_allowed_arguments())

        unallowed = set(self.request.query_params.keys()) - allowed_params

        if unallowed:
            msg = 'Unsupported parameter(s): {}. Please use a combination of: {}.'.format(
                ', '.join(unallowed),
                ', '.join(allowed_params),
            )
            self.form.add_error(field=None, error=ParseError(msg))

    def is_valid(self):
        """Validate filterset."""
        self.validate_query_params()
        return super().is_valid()


class BaseResolweFilter(CheckQueryParamsMixin, filters.FilterSet):
    """Base filter for Resolwe's endpoints."""

    class Meta:
        """Filter configuration."""

        fields = {
            'id': NUMBER_LOOKUPS[:],
            'slug': TEXT_LOOKUPS[:],
            'name': TEXT_LOOKUPS[:],
            'contributor': ['exact', 'in'],
            'created': DATETIME_LOOKUPS[:],
            'modified': DATETIME_LOOKUPS[:],
        }


class DescriptorSchemaFilter(BaseResolweFilter):
    """Filter the DescriptorSchema endpoint."""

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = DescriptorSchema


class CollectionFilter(BaseResolweFilter):
    """Filter the Collection endpoint."""

    data = filters.ModelChoiceFilter(queryset=Data.objects.all())
    entity = filters.ModelChoiceFilter(queryset=Entity.objects.all())

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Collection
        fields = {**BaseResolweFilter.Meta.fields, **{
            'description': TEXT_LOOKUPS[:],
            'descriptor_schema': ['exact'],
        }}


class TagsFilter(filters.filters.BaseCSVFilter, filters.CharFilter):
    """Filter for tags."""

    def __init__(self, *args, **kwargs):
        """Construct tags filter."""
        kwargs.setdefault('lookup_expr', 'contains')
        super().__init__(*args, **kwargs)


class EntityFilter(CollectionFilter):
    """Filter the Entity endpoint."""

    collection = filters.ModelChoiceFilter(field_name='collection', queryset=Collection.objects.all())
    descriptor_completed = filters.rest_framework.filters.BooleanFilter(field_name='descriptor_completed')
    tags = TagsFilter()

    class Meta(CollectionFilter.Meta):
        """Filter configuration."""

        model = Entity


class ProcessFilter(BaseResolweFilter):
    """Filter the Process endpoint."""

    category = filters.CharFilter(field_name='category', lookup_expr='startswith')
    type = filters.CharFilter(field_name='type', lookup_expr='startswith')
    is_active = filters.rest_framework.filters.BooleanFilter(field_name='is_active')

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Process
        fields = {**BaseResolweFilter.Meta.fields, **{
            'scheduling_class': ['exact'],
        }}


class CharInFilter(filters.BaseInFilter, filters.CharFilter):
    """Helper class for creation of CharFilter with "in" lookup."""


class DataFilter(BaseResolweFilter):
    """Filter the Data endpoint."""

    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    collection__slug = filters.CharFilter(field_name='collection__slug', lookup_expr='exact')

    entity = filters.ModelChoiceFilter(queryset=Entity.objects.all())

    type = filters.CharFilter(field_name='process__type', lookup_expr='startswith')
    status = filters.CharFilter(lookup_expr='iexact')
    status__in = CharInFilter(field_name='status', lookup_expr='in')

    tags = TagsFilter()

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Data
        fields = {**BaseResolweFilter.Meta.fields, **{
            'process': ['exact'],
            'process__slug': ['exact'],
            'finished': DATETIME_LOOKUPS[:],
            'started': DATETIME_LOOKUPS[:],
        }}


class RelationFilter(BaseResolweFilter):
    """Filter the Relation endpoint."""

    category = filters.CharFilter(lookup_expr='iexact')
    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    type = filters.CharFilter(field_name='type__name')

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Relation
        fields = BaseResolweFilter.Meta.fields

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return super().get_always_allowed_arguments() + (
            'entity',
            'label',
            'position',
        )



# pylint: disable=missing-docstring
import os

from resolwe.flow.models import Data
from resolwe.test import ProcessTestCase, tag_process, with_docker_executor


class DirsProcessTestCase(ProcessTestCase):

    @with_docker_executor
    def setUp(self):
        super().setUp()
        self.files_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'files'))

        self.dir = self.run_process('upload-dir', {'src': 'compressed dir.tar.gz'})
        self.dir_sym_link = self.run_process('upload-dir', {'src': 'dir sym link.tar.gz'})

    @tag_process('upload-dir')
    def test_assert_dir_exists(self):
        self.assertDirExists(self.dir, 'dir')
        self.assertDirExists(self.dir_sym_link, 'dir')

    @tag_process('upload-dir')
    def test_assert_dir(self):
        self.assertDir(self.dir, 'dir', 'compressed dir.tar.gz')
        self.assertDir(self.dir_sym_link, 'dir', 'dir sym link.tar.gz')

    @tag_process('upload-dir')
    def test_assert_dir_structure(self):
        # Complete tree structure of 'compressed dir.tar.gz'
        structure = {
            'file binary': None,
            'text file.txt': None,
            'root dir': {
                'file binary': None,
                'text file.txt': None,
                'nested dir.ext': {
                    'empty dir': {
                    },
                },
            },
        }

        partial_structure = {
            'root dir': {
                'file binary': None,
            },
        }

        sym_link = {
            'text file.txt': None,
            'sym link': None,
        }

        self.assertDirStructure(self.dir, 'dir', structure)
        self.assertRaisesRegex(
            AssertionError, r'Directory structure mismatch \(exact check\).',
            ProcessTestCase.assertDirStructure, self, self.dir, 'dir', partial_structure)
        self.assertDirStructure(self.dir, 'dir', partial_structure, exact=False)

        self.assertDirStructure(self.dir_sym_link, 'dir', sym_link)

    @with_docker_executor
    @tag_process('upload-dir')
    def test_bad_dir(self):
        self.run_process('upload-dir', {'src': 'dir bad format.tar.gz'}, Data.STATUS_ERROR)
